import 'dart:async';
import 'dart:io';

import 'package:dynatrace_flutter_plugin/dynatrace_flutter_plugin.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';

import 'model/configuration.dart';
import 'http/http_instrumentation.dart';
import 'interface/agent_interface.dart';
import 'action/null_actions/null_root_action_impl.dart';
import 'model/data_collection_level.dart';
import 'util/logger.dart';
import 'util/string_utils.dart';
import 'action/action.dart';
import 'action/counter.dart';
import 'action/interface/root_action.dart';
import 'model/platform.dart';

class DynatraceImpl implements Dynatrace {
  static Dynatrace _instance;
  final MethodChannel _methodChannel;

  final Function _startupUpFunction;

  bool _started = false;
  bool _crashReporting = true;

  factory DynatraceImpl() {
    if (_instance == null) {
      final MethodChannel _methodChannel =
          const MethodChannel('dynatrace_flutter_plugin/dynatrace');
      _instance = DynatraceImpl.private(_methodChannel);
    }

    return _instance;
  }

  DynatraceImpl.private(this._methodChannel,
      {Function startUpFunction = runApp})
      : _startupUpFunction = startUpFunction;

  /// Returns if the plugin is already enabled or not. This depends if
  /// there is a configuration available which will only be set in the start
  /// method.
  bool _isEnabled() {
    return _started;
  }

  Future<void> start(Widget topLevelWidget,
      {Configuration configuration = const Configuration()}) async {
    FlutterError.onError = (FlutterErrorDetails details) async {
      FlutterError.dumpErrorToConsole(details, forceReport: true);
      if (_crashReporting && details.exception.runtimeType != null) {
        String errorName = details.exception.runtimeType.toString();
        String reason = details.toStringShort() != null
            ? "${details.exceptionAsString()} - ${details.toStringShort()}"
                .substring(errorName.length + 2)
            : details.exceptionAsString();

        await reportCrash(errorName, reason, details.stack.toString());
      }
    };

    await runZoned(() async {
      await _start(topLevelWidget, configuration: configuration);
    }, onError: (error, stacktrace) async {
      if (_crashReporting && error.runtimeType != null) {
        await reportCrash(error.runtimeType.toString(),
            error.message.toString(), stacktrace.toString());
      }
    });
  }

  Future<void> _start(Widget topLevelWidget,
      {Configuration configuration = const Configuration()}) async {
    WidgetsFlutterBinding.ensureInitialized();
    Map<String, bool> autoStartConfiguration =
        await _methodChannel.invokeMapMethod("getAutoStartConfiguration");

    if (!_isEnabled()) {
      _started = true;

      if (autoStartConfiguration["autoStart"]) {
        // Take over the log level of native agent
        configuration = Configuration(
            logLevel: autoStartConfiguration["logLevel"]
                ? LogLevel.Debug
                : LogLevel.Info,
            reportCrash: autoStartConfiguration["crashReporting"],
            monitorWebRequest: autoStartConfiguration["webRequest"]);
      } else {
        if (configuration.isAutoStartPropertyAvailable) {
          // We need to do a manual startup if it is available
          _methodChannel.invokeMethod(
              "start", configuration.getStartupConfiguration());
        }
      }

      Logger().logLevel = configuration.logLevel;

      if (configuration.isWebRequestMonitoringEnabled) {
        // Override the HTTP communication with our handler
        HttpOverrides.global = new DynatraceHttpOverrides(_instance);
      } else {
        Logger().d("Webrequest Instrumentation deactivated.",
            logType: LogType.Info);
      }

      _crashReporting = configuration.isCrashReportingEnabled;

      if (!configuration.isCrashReportingEnabled) {
        Logger().d("Crash reporting deactivated.", logType: LogType.Info);
      }
    }

    Logger().i("Dynatrace Flutter Plugin started!", logType: LogType.Info);

    _startupUpFunction(topLevelWidget);
  }

  DynatraceRootAction enterAction(String name, {Platform platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace().start() method was not called! Action will have no effect.",
          logType: LogType.Warning);
      return DynatraceNullRootAction();
    }

    if (StringUtils.isStringNullOrEmpty(name)) {
      Logger().d("Action Name was empty or null! Action will have no effect.",
          logType: LogType.Error);
      return DynatraceNullRootAction();
    }

    int key = ActionCounter().getNewActionId();
    _methodChannel.invokeMethod("enterAction", <String, dynamic>{
      "name": name,
      "key": key,
      "platform": platform != null ? platform.index : null
    });

    return DynatraceRootActionImpl.private(key, _methodChannel);
  }

  Future<String> getHTTPTagForWebRequest(String url) async {
    return await _methodChannel
        .invokeMethod("getRequestTag", <String, dynamic>{"url": url});
  }

  Future<int> startWebRequestTiming(String requestTag, String url) async {
    if (!StringUtils.isStringNullEmptyOrWhitespace(requestTag) &&
        !StringUtils.isStringNullEmptyOrWhitespace(url)) {
      return await _methodChannel.invokeMethod("startWebRequestTiming",
          <String, dynamic>{"requestTag": requestTag, "url": url});
    } else {
      Logger().d(
          "Web Request Timing could not be created! Url: $url requestTag: $requestTag",
          logType: LogType.Error);
    }

    return -1;
  }

  Future<bool> stopWebRequestTiming(int webTimingKey, String url,
      int responseCode, String responseMessage) async {
    if (webTimingKey != null &&
        webTimingKey != -1 &&
        !StringUtils.isStringNullEmptyOrWhitespace(url) &&
        responseCode != null &&
        !StringUtils.isStringNullEmptyOrWhitespace(responseMessage)) {
      return await _methodChannel
          .invokeMethod("stopWebRequestTiming", <String, dynamic>{
        "webTimingKey": webTimingKey,
        "url": url,
        "responseCode": responseCode,
        "responseMessage": responseMessage
      });
    } else {
      Logger().d(
          "Web Request Timing could not be stopped!" +
              "- WebTimingKey: $webTimingKey Url: $url responseCode: $responseCode responseMessage: $responseMessage",
          logType: LogType.Error);
    }

    return false;
  }

  void reportError(String errorName, int errorCode, {Platform platform}) {
    if (!_isEnabled()) {
      Logger().i("Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    if (!StringUtils.isStringNullOrEmpty(errorName) && errorCode != null) {
      _methodChannel.invokeMethod("reportError", <String, dynamic>{
        "errorName": errorName,
        "errorCode": errorCode,
        "platform": platform != null ? platform.index : null
      });
    } else {
      Logger().d(
          "Error was not reported! ErrorName:$errorName errorCode:$errorCode",
          logType: LogType.Error);
    }
  }

  Future<void> reportCrash(String errorName, String reason, String stacktrace,
      {Platform platform}) async {
    if (!_isEnabled()) {
      Logger().i("Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    if (!StringUtils.isStringNullOrEmpty(errorName)) {
      if (_crashReporting) {
        await _methodChannel.invokeMethod("reportCrash", <String, dynamic>{
          "errorValue": errorName,
          "reason": reason,
          "stacktrace": stacktrace,
          "platform": platform != null ? platform.index : null
        });
      }
    } else {
      Logger().d(
          "Crash will not be reported because something is missing! errorName:$errorName reason:$reason stacktrace: $stacktrace",
          logType: LogType.Error);
    }
  }

  void endSession({Platform platform}) {
    if (_isEnabled()) {
      _methodChannel.invokeMethod("endVisit", <String, dynamic>{
        "platform": platform != null ? platform.index : null
      });
    } else {
      Logger().i("Dynatrace().start() method was not called!",
          logType: LogType.Warning);
    }
  }

  void identifyUser(String user, {Platform platform}) {
    if (!_isEnabled()) {
      Logger().i("Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    _methodChannel.invokeMethod("identifyUser", <String, dynamic>{
      "user": user,
      "platform": platform != null ? platform.index : null
    });
  }

  void setGPSLocation(double latitude, double longitude, {Platform platform}) {
    if (!_isEnabled()) {
      Logger().i("Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    if (latitude != null && longitude != null) {
      _methodChannel.invokeMethod("setGPSLocation", <String, dynamic>{
        "latitude": latitude,
        "longitude": longitude,
        "platform": platform != null ? platform.index : null
      });
    } else {
      Logger().d(
          "Position was not reported because of missing values! Latitidue:$latitude Longitude:$longitude",
          logType: LogType.Error);
    }
  }

  Future<bool> isCrashReportingOptedIn({Platform platform}) async {
    return await _methodChannel.invokeMethod(
        "isCrashReportingOptedIn", <String, dynamic>{
      "platform": platform != null ? platform.index : null
    });
  }

  void setCrashReportingOptedIn(bool crashReporting, {Platform platform}) {
    if (!_isEnabled()) {
      Logger().i("Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    if (crashReporting != null) {
      _methodChannel.invokeMethod("setCrashReportingOptedIn", <String, dynamic>{
        "crashReporting": crashReporting,
        "platform": platform != null ? platform.index : null
      });
    } else {
      Logger().d("Crash reporting not set because it is null",
          logType: LogType.Error);
    }
  }

  Future<DataCollectionLevel> getDataCollectionLevel(
      {Platform platform}) async {
    return DataCollectionLevel.values[await _methodChannel.invokeMethod(
        "getDataCollectionLevel", <String, dynamic>{
      "platform": platform != null ? platform.index : null
    })];
  }

  void setDataCollectionLevel(DataCollectionLevel dataCollectionLevel,
      {Platform platform}) {
    if (!_isEnabled()) {
      Logger().i("Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    if (dataCollectionLevel != null) {
      _methodChannel.invokeMethod("setDataCollectionLevel", <String, dynamic>{
        "dataCollectionLevel": dataCollectionLevel.index,
        "platform": platform != null ? platform.index : null
      });
    } else {
      Logger().d("DataCollection was not set because it is null!",
          logType: LogType.Error);
    }
  }

  void flushEvents({Platform platform}) {
    if (_isEnabled()) {
      _methodChannel.invokeMethod("flushEvents", <String, dynamic>{
        "platform": platform != null ? platform.index : null
      });
    } else {
      Logger().i("Dynatrace().start() method was not called!",
          logType: LogType.Warning);
    }
  }
}
